# pylint: disable=missing-docstring,line-too-long

import unittest
from json import dumps
from typing import Dict, NamedTuple, Optional
from unittest.mock import MagicMock, patch

import soil


class MockHttpResponse(NamedTuple):
    """Soil configuration class"""

    status_code: int
    text: str


# pylint: disable=unused-argument
def mock_http_post(
    url: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Dict[str, str]] = None,
    timeout: int,
) -> MockHttpResponse:
    assert json is not None
    response = "ok"
    return MockHttpResponse(status_code=201, text=dumps(response))


class TestAlerts(unittest.TestCase):
    @patch("soil.api.session.post", side_effect=mock_http_post)
    def test_create_event(self, mock_post: MagicMock) -> None:
        soil.alerts.event("test_key", "test_value")
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/alerts/events/",
            json={"key": "test_key", "value": "test_value"},
            timeout=30,
        )

    @patch("soil.api.session.post", side_effect=mock_http_post)
    def test_create_alert(self, mock_post: MagicMock) -> None:
        alert_config = {
            "role": "test_alerts",
            "event_id": "test_event1",
            "condition": {"type": "any"},
        }
        soil.alerts.alert(alert_config)
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/alerts/alerts/",
            json=alert_config,
            timeout=30,
        )



"""
Asynchronous Soil package.
"""

from soil.aio import alerts, api
from soil.aio.alias import alias
from soil.aio.data import data
from soil.aio.dictionary import dictionary
from soil.aio.job import job

__all__ = [
    "data",
    "alias",
    "alerts",
    "dictionary",
    "job",
    "api",
]



"""Exports the soil logger"""

import importlib
import logging
import logging.handlers
import os
import sqlite3
import sys
from os.path import exists
from typing import LiteralString, cast

from psycopg_pool import ConnectionPool

from soil.types import TypeLog

_SOIL_LOGGER = "soil_logger"
TABLE_NAME = "processed_files"
PROCESSED_FILES_DB = "/rain-data/processed_files.db"
DB_POOL = None


if not exists(PROCESSED_FILES_DB):
    PROCESSED_FILES_DB = None

try:
    sys.path.insert(1, "/rain/config")
    config = importlib.import_module(os.environ["PY_ENV"])
    if config.db_credentials is not None:
        DB_POOL = ConnectionPool(
            " ".join([f"{k}={v}" for k, v in config.db_credentials.items()])
        )
except (ModuleNotFoundError, AttributeError, KeyError):
    pass


def _get_connection():
    if DB_POOL is not None:
        return DB_POOL.connection()
    return sqlite3.connect(PROCESSED_FILES_DB)


def _transform_query(query: str) -> str:
    if DB_POOL is None:
        return query.replace("%s", "?")
    return query


logger = logging.getLogger(_SOIL_LOGGER)  # pylint: disable=invalid-name


def logger_extra_kwarg(*, type_log: TypeLog, file_name: str) -> dict[str, str]:
    """Creates the extra kwarg for the logger."""
    return {
        HashFileHandler.type_log: type_log.value,
        HashFileHandler.hashfile: file_name,
    }


def set_file_status(*, status: TypeLog, file_hash: str, message_status: str) -> None:
    """Updates the status of the file hash."""
    logger.info(
        message_status, extra=logger_extra_kwarg(type_log=status, file_name=file_hash)
    )


class HashFileHandler(logging.StreamHandler):
    """Handler of logs to store file to sqlite."""

    type_log = "type"
    hashfile = "file"
    state_name_column = "state_description"
    message_column = "message"

    def emit(self, record: logging.LogRecord) -> None:
        """Emits the log record."""
        type_log = cast(TypeLog | None, getattr(record, "type", None))
        hashfile = cast(str | None, getattr(record, "file", None))
        if (
            type_log is not None
            and hashfile is not None
            and type_log in [*TypeLog]
            and (PROCESSED_FILES_DB is not None or DB_POOL is not None)
        ):
            self._process_file_in_storage(
                hashfile=hashfile, type_log=type_log, message=record.getMessage()
            )

    def _process_file_in_storage(self, **kwargs) -> None:
        """Process the file hash to update the DB."""
        with _get_connection() as connection:
            cursor = connection.cursor()
            self._update_hashfile(cursor=cursor, table_name=TABLE_NAME, **kwargs)

    def _update_hashfile(
        self,
        *,
        cursor: sqlite3.Cursor,
        table_name: LiteralString,
        hashfile: str,
        type_log: TypeLog,
        message: str,
    ) -> None:
        cursor.execute(
            _transform_query(
                f"UPDATE {table_name} "  # nosec
                f"SET {self.state_name_column} = %s, {self.message_column} = %s "
                f"WHERE file_name = %s and "
                f"({self.state_name_column} IS NULL or {self.state_name_column} = %s)"
            ),
            (type_log, message, hashfile, TypeLog.NOT_PROCESSED.value),
        )


logger.addHandler(HashFileHandler())



"""Package that contains calls for the alerts system."""

from typing import Any, Dict

from soil.aio.api import create_alert, create_event


async def event(key: str, value: Any) -> Any:
    """Creates an event asynchronously."""
    return await create_event(key, value)


async def alert(alert_config: Dict) -> Any:
    """Creates an alert asynchronously."""
    return await create_alert(alert_config)



"""Module that implements SOIL's Command LIne Interface"""

import argparse
import getpass
import json
import os
import shutil
import subprocess  # nosec
import sys
from json import JSONDecodeError
from typing import Dict, Optional, Tuple

import requests
import yaml

parser = argparse.ArgumentParser(prog="soil")
subparsers = parser.add_subparsers(dest="command")

login_cmd = subparsers.add_parser("login", help="Login to configured soil instance")
login_cmd.add_argument("--user", type=str, help="User name or email of the user")
login_cmd.add_argument("--password", type=str, help="Password")
login_cmd.add_argument("--trust", type=str, required=False, help="2FA trust cookie")

run_cmd = subparsers.add_parser(
    "run", help="Runs the provided module in the current soil project"
)
run_cmd.add_argument(
    "chapter",
    metavar="chapter",
    help="Chapter in soil.yml (setup, data, migration, ...)",
)
run_cmd.add_argument("module", metavar="module", help="Module of chapter in soil.yml")
run_cmd.add_argument(
    "module_args", nargs=argparse.REMAINDER, help="Arguments to be passed to the module"
)


def _save_environment(config: Dict, env: Dict) -> None:
    """Saves the environment keys to a file.
    The file is located in ~/.soil/soil.env and ~/.soil/:app_id/soil.env
    Where app_id is the app_id in the config.
    Notice that ~/.soil/soil.env will store the last login only.
    """
    try:
        os.makedirs(os.path.expanduser("~/.soil/" + config["auth_app_id"]))
    except FileExistsError:
        # directory already exists
        pass

    with open(
        os.path.expanduser("~/.soil/soil.env"), "w", encoding="utf-8"
    ) as env_file:
        env_file.write(json.dumps(env, indent=4, sort_keys=True))

    shutil.copyfile(
        os.path.expanduser("~/.soil/soil.env"),
        os.path.expanduser("~/.soil/" + config["auth_app_id"] + "/soil.env"),
    )


def _get_soil_root(relpath: str):  # type: ignore
    """Checks if the current dir is under a soil environment
    and returns its root. Returns None otherwise.
    """
    path = os.path.abspath(relpath) + "/"

    while path != "/":
        path, _ = os.path.split(path)
        if "soil.yml" in os.listdir(path):
            return path

    return None


def _soil_init() -> Tuple[Dict, Optional[Dict]]:
    """Loads configuration and environment."""
    project_root = _get_soil_root(".")
    try:
        if project_root is None:
            raise FileNotFoundError("Project root not found.")
        with open(project_root + "/soil.conf", encoding="utf-8") as config_file:
            config = json.loads(config_file.read())
    except FileNotFoundError:
        try:
            with open(
                os.path.expanduser("~/.soil/soil.conf"), "r", encoding="utf-8"
            ) as config_file:
                config = json.loads(config_file.read())
        except (IOError, JSONDecodeError):
            if sys.argv[1] != "configure":
                try:
                    os.rename(
                        os.path.expanduser("~/.soil/soil.conf.bak"),
                        os.path.expanduser("~/.soil/soil.conf"),
                    )
                except IOError:
                    print(
                        "Can not load soil configuration. Plase run soil configure to configure it."
                    )
                    sys.exit(1)
            else:
                config = None
    try:
        with open(
            os.path.expanduser("~/.soil/soil.env"), "r", encoding="utf-8"
        ) as env_file:
            env = json.loads(env_file.read())
        return config, env
    except (IOError, JSONDecodeError):
        if sys.argv[1] != "configure" and sys.argv[1] != "login":
            try:
                os.rename(
                    os.path.expanduser("~/.soil/soil.conf.bak"),
                    os.path.expanduser("~/.soil/soil.conf"),
                )
            except IOError:
                print(
                    "Can not load soil environment. Plase run soil configure to initialize it."
                )
                sys.exit(1)
        return config, None


def _login(args: argparse.Namespace, config: Dict, env: Optional[Dict]) -> None:
    """Authenticates to the authentication backend and stores the credentials (JWT) in the environment"""
    print(f"Authenticating to {config['auth_url']} for app {config['auth_app_id']}...")

    if env is None:
        env = {}

    if args.user is not None and args.password is not None:
        username = args.user
        password = args.password
    else:
        if "auth" in env:
            username = input(
                "Username: ["
                + env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
                + "]"
            )  # nosec - Input is safe in python3
            if username == "":
                username = env["auth"]["user"].get(
                    "username", env["auth"]["user"].get("email", "")
                )
        else:
            username = input("Username: ")  # nosec - Input is safe in python3
        password = getpass.getpass()

    request_json = {
        "loginId": username,
        "password": password,
        "applicationId": config["auth_app_id"],
    }

    resp = requests.post(
        config["auth_url"] + "/api/login",
        headers={"Authorization": config["auth_api_key"]},
        json=request_json,
        timeout=10,
        cookies={"fusionauth.trust": args.trust} if args.trust is not None else None,
    )

    if resp.status_code in {200, 202}:
        env["auth"] = json.loads(resp.content)
        _save_environment(config, env)
        print(
            "Successfully logged in as "
            + env["auth"]["user"].get("username", env["auth"]["user"].get("email", ""))
            + "!"
        )
    elif resp.status_code == 404:
        print("The user was not found or the password was incorrect.")
        sys.exit(1)
    elif resp.status_code == 400:
        print(
            f"Invalid parameters. Perhaps the api key {config['auth_api_key']} or \
                 the app id {config['auth_app_id']} are not correct."
        )
        sys.exit(1)
    else:
        print("Login failed with status code", resp.status_code)
        sys.exit(1)


def _run(args: argparse.Namespace) -> None:
    """Runs the script provided as argument using the python in the virtual environment"""
    soil_root = _get_soil_root(".")
    if not soil_root:
        print(
            "This folder is not initalized as a soil project. "
            "Please run soil init to initialize it."
        )
        sys.exit(1)

    chapter = vars(args)["chapter"]
    module = vars(args)["module"]
    module_args = vars(args)["module_args"]

    os.chdir(soil_root)
    with open("soil.yml", encoding="utf-8") as f:
        conf = yaml.safe_load(f)
    try:
        script = next(
            script for script in conf[chapter] if script.get(module) is not None
        )[module]
    except (IndexError, KeyError):
        print("Script", module, "not found.")
        sys.exit(1)
    params = script.get("params", {})
    params = [["--" + k, str(v)] for (k, v) in params.items()]
    params = [item for sublist in params for item in sublist]
    try:
        shell = [".venv/bin/python", "-m", script["path"], *params, *module_args]
        print("Running:", " ".join(shell), flush=True)
        subprocess.run(shell, check=True)  # nosec
        print("Finished:", " ".join(shell), flush=True)
    except subprocess.CalledProcessError:
        print("Failed executing:", " ".join(shell), flush=True)
        sys.exit(1)


def main() -> None:
    """Main function"""
    args = parser.parse_args()

    config, env = _soil_init()

    try:
        if args.command == "login":
            _login(args, config, env)
        elif args.command == "run":
            _run(args)
    except KeyboardInterrupt:
        print("\nKeyboardInterrupt - Exit...")
        try:
            sys.exit(0)
        except SystemExit:
            os._exit(0)  # pylint: disable=protected-access


if __name__ == "__main__":
    main()



"""
Soil package.
"""

from soil import aio, connectors, errors, finder, types
from soil.alerts import alerts
from soil.alias import alias
from soil.data import data
from soil.data_structure import DataStructure
from soil.decorator import decorator
from soil.dictionary import dictionary
from soil.job import job
from soil.logger import logger, logger_extra_kwarg, set_file_status
from soil.modulify import modulify
from soil.task import task, task_wait

finder.upload_modules()

__all__ = [
    "modulify",
    "data_structures",
    "DataStructure",
    "modules",
    "data",
    "alias",
    "logger",
    "connectors",
    "decorator",
    "task",
    "task_wait",
    "alerts",
    "dictionary",
    "errors",
    "types",
    "logger_extra_kwarg",
    "set_file_status",
    "job",
    "aio",
]



import unittest
from unittest.mock import AsyncMock, patch

from soil.aio.job import job
from soil.data_structure import DataStructure
from soil.job import _Job
from soil.pipeline import Pipeline


class TestJob(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.mock_experiment = {
            "_id": "mock_experiment_id",
            "outputs": {"mock_sym_id": "mock_result_id"},
            "created_at": 1609459200000,  # 2021-01-01 00:00:00 UTC
        }

    def create_data_structure(self):
        pipeline = Pipeline()
        pipeline.plan = "mock_plan"  # pyright:ignore[reportAttributeAccessIssue]
        return DataStructure(sym_id="mock_sym_id", pipeline=pipeline)

    async def test_job_function(self):
        data_structure = self.create_data_structure()

        with patch(
            "soil.aio.job.create_experiment", new_callable=AsyncMock
        ) as mock_create_experiment:
            mock_create_experiment.return_value = self.mock_experiment

            result = await job(
                data_structure, group="test_group", description="test_description"
            )

            self.assertIsInstance(result, _Job)
            self.assertEqual(result.experiment_id, "mock_experiment_id")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result.result_id, "mock_result_id")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result._created_at, "2021-01-01T00:00:00+00:00")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result._group, "test_group")  # pyright:ignore[reportAttributeAccessIssue]

            mock_create_experiment.assert_called_once_with(
                "mock_plan",
                experiment_group="test_group",
                description="test_description",
            )

    async def test_job_function_default_description(self):
        data_structure = self.create_data_structure()

        with patch(
            "soil.aio.job.create_experiment", new_callable=AsyncMock
        ) as mock_create_experiment:
            mock_create_experiment.return_value = self.mock_experiment

            result = await job(data_structure, group="test_group")

            self.assertIsInstance(result, _Job)
            self.assertEqual(result.experiment_id, "mock_experiment_id")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result.result_id, "mock_result_id")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result._created_at, "2021-01-01T00:00:00+00:00")  # pyright:ignore[reportAttributeAccessIssue]
            self.assertEqual(result._group, "test_group")  # pyright:ignore[reportAttributeAccessIssue]

            mock_create_experiment.assert_called_once_with(
                "mock_plan", experiment_group="test_group", description=None
            )

    async def test_job_function_missing_pipeline(self):
        data_structure = DataStructure(sym_id="mock_sym_id")

        with self.assertRaises(ValueError) as context:
            await job(data_structure, group="test_group")

        self.assertEqual(str(context.exception), "Pipeline plan is required")

    async def test_job_function_missing_sym_id(self):
        pipeline = Pipeline()
        pipeline.plan = "mock_plan"  # pyright:ignore[reportAttributeAccessIssue]
        data_structure = DataStructure(pipeline=pipeline)

        with self.assertRaises(ValueError) as context:
            await job(data_structure, group="test_group")

        self.assertEqual(str(context.exception), "sym_id is required")


if __name__ == "__main__":
    unittest.main()



# pylint:disable=missing-module-docstring,missing-class-docstring,missing-function-docstring
# pylint:disable=import-outside-toplevel
import logging
import sqlite3
import unittest
from pathlib import Path
from unittest.mock import Mock, call, patch

from soil.logger import set_file_status
from soil.types import TypeLog

_TEST_LOGGER = "test_logger"
PROCESSED_FILES_DB = "test/processed_files.db"


@patch("soil.logger.PROCESSED_FILES_DB", new=PROCESSED_FILES_DB)
@patch("soil.logger._SOIL_LOGGER", new=_TEST_LOGGER)
class TestHashfileLogger(unittest.TestCase):
    @staticmethod
    def _init_storage():
        """Checks if the state files are writeable."""
        path = Path(PROCESSED_FILES_DB)
        if not path.parent.exists():
            path.mkdir(parents=True, exist_ok=True)

        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            connection.execute("DROP TABLE IF EXISTS processed_files")
            connection.execute(
                """
                CREATE TABLE processed_files (
                    file_name VARCHAR PRIMARY KEY,
                    date VARCHAR NOT NULL,
                    size int NOT NULL,
                    datatype VARCHAR NOT NULL,
                    state VARCHAR CHECK( state IN ('WAITING','PROCESSED','ERROR') ) NOT NULL,
                    state_description VARCHAR,
                    message VARCHAR
                )
                """
            )

    @staticmethod
    def _store_file_db(hexdigest, date, size, datatype):
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            connection.execute(
                """
                INSERT INTO processed_files (file_name, date, size, datatype, state)
                VALUES (?,?,?,?,?)
                """,
                (hexdigest, date, size, datatype, "WAITING"),
            )

    def test_logs_handler(self):
        # importing inside method because of patch test_logger
        from soil.logger import logger, logger_extra_kwarg

        TestHashfileLogger._init_storage()
        file = "hash_file_1"
        file_2 = "hash_file_2"
        file_3 = "hash_file_3"
        TestHashfileLogger._store_file_db(file, "2022-02-02", 1, "test1")
        TestHashfileLogger._store_file_db(file_2, "2022-02-02", 1, "test1")
        TestHashfileLogger._store_file_db(file_3, "2022-02-02", 1, "test1")
        logger.setLevel(logging.DEBUG)
        logger.debug(
            "%s has been processed successfully!",
            file,
            extra=logger_extra_kwarg(type_log=TypeLog.PROCESSED, file_name=file),
        )
        logger.debug(
            "%s has been processed successfully!",
            "file_not_existent",
            extra=logger_extra_kwarg(
                type_log=TypeLog.PROCESSED, file_name="file_not_existent"
            ),
        )
        logger.debug(
            "%s data not consistent!",
            file_2,
            extra=logger_extra_kwarg(type_log=TypeLog.NOT_CONSISTENT, file_name=file_2),
        )
        logger.debug(
            "%s has been processed successfully!",
            file_2,
            extra=logger_extra_kwarg(type_log=TypeLog.PROCESSED, file_name=file_2),
        )
        logger.debug(
            "%s error",
            file_3,
            extra=logger_extra_kwarg(type_log=TypeLog.NOT_PROCESSED, file_name=file_3),
        )
        logger.debug(
            "%s processed!",
            file_3,
            extra=logger_extra_kwarg(type_log=TypeLog.PROCESSED, file_name=file_3),
        )
        with sqlite3.connect(PROCESSED_FILES_DB) as connection:
            results = connection.execute(
                """
                    SELECT *
                    FROM processed_files
                """
            ).fetchall()
        self.assertListEqual(
            [
                (
                    file,
                    "2022-02-02",
                    1,
                    "test1",
                    "WAITING",
                    "processed",
                    f"{file} has been processed successfully!",
                ),
                (
                    file_2,
                    "2022-02-02",
                    1,
                    "test1",
                    "WAITING",
                    "not_consistent",
                    f"{file_2} data not consistent!",
                ),
                (
                    file_3,
                    "2022-02-02",
                    1,
                    "test1",
                    "WAITING",
                    "processed",
                    f"{file_3} processed!",
                ),
            ],
            results,
        )

    @patch("soil.logger.logger")
    def test_set_file_status(self, mocked_logger: Mock) -> None:
        set_file_status(
            status=TypeLog.NOT_PROCESSED, file_hash="123", message_status="test_message"
        )
        self.assertListEqual(
            mocked_logger.info.call_args_list,
            [call("test_message", extra={"type": "not_processed", "file": "123"})],
        )



# pylint: disable=missing-docstring
import unittest
from unittest.mock import AsyncMock, Mock, call, patch

from soil.finder import _upload_modules


def _mock_file_generator(file_name: str, folder: str = "") -> Mock:
    mock = Mock(name=file_name, stem=file_name.split(".")[0])
    mock.read_text.return_value = file_name
    mock.__str__ = lambda _: f"{folder}/{file_name}"
    return mock


class TestFinder(unittest.IsolatedAsyncioTestCase):
    @patch("soil.finder.api", new_callable=AsyncMock)
    @patch("soil.finder.Path")
    async def test_upload_modules(self, mock_path: Mock, mock_api: Mock):
        mock_api.get_modules.return_value = [
            {"name": "apis.test1", "hash": "1111"},
            {"name": "apis.test2", "hash": "2222"},
            {
                "name": "apis.test.apis_file1",
                "hash": "e836da06f84039b439725d2dc461aa21a75daa06852f41505c21ae9cfd46bec3",
            },
            {
                "name": "apis.test",
                "hash": "8e2c1f727276d30ced72fa67cb848772a9802162bec39b42a8df5359fd063a4c",
            },
            {"name": "arum.arum_file1", "hash": "5555"},
        ]
        venv = Mock()
        venv.name = ".venv"
        venv.is_dir.return_value = True
        venv.rglob.return_value = [_mock_file_generator("venv_file1.py")]
        test = Mock()
        test.name = "test"
        test.is_dir.return_value = True
        test.rglob.return_value = [_mock_file_generator("test_file1.py")]
        file_py = Mock()
        file_py.name = "abc.py"
        file_py.is_dir.return_value = False
        apis = Mock()
        apis.name = "apis"
        apis.is_dir.return_value = True
        apis.rglob.return_value = [
            _mock_file_generator("apis_file1.py", "apis/test"),
            _mock_file_generator("__init__.py", "apis/test"),
        ]
        arum = Mock()
        arum.name = "arum"
        arum.is_dir.return_value = True
        arum.rglob.return_value = [
            _mock_file_generator("arum_file1.py", "arum"),
            _mock_file_generator("__init__.py", "arum"),
        ]

        mock_path().glob.return_value = [venv, test, file_py, apis, arum]

        await _upload_modules()
        self.assertListEqual(
            mock_api.upload_modules.call_args_list,
            [
                call(
                    modules=[
                        {
                            "name": "arum.arum_file1",
                            "code": "arum_file1.py",
                            "is_package": False,
                        },
                        {"name": "arum", "code": "__init__.py", "is_package": True},
                    ]
                )
            ],
        )



"""Package that contains async calls to the SOIL's REST API"""

import gzip
import json
import logging
import os
from contextlib import asynccontextmanager
from itertools import takewhile
from typing import Any, Dict, Literal, TypedDict
from urllib.parse import urlencode

import httpx

from soil import errors
from soil.configuration import CONF, GLOBAL_CONFIG
from soil.types import (
    Context,
    CreateGraphLogRequest,
    Experiment,
    GetModule,
    GetModuleHash,
    GraphStateResponse,
    PipelineForRequest,
    Plan,
    SaveGraphStateRequest,
)

logger = logging.getLogger(__name__)

# Configure the logger based on the SOIL_SDK_LOGGER environment variable
log_level = os.getenv("SOIL_SDK_LOGGER", "WARNING").upper()
logger.setLevel(log_level)

API_ROOT = f"{GLOBAL_CONFIG.host!s}/v2/"
API_ROOT_V3 = f"{GLOBAL_CONFIG.host!s}/v3/"

_client: httpx.AsyncClient | None = None


async def clean_client():
    """Closes and cleans up the global httpx client if it exists."""
    global _client  # noqa: PLW0603
    if _client is not None:
        await _client.aclose()
        _client = None


@asynccontextmanager
async def _get_client():
    """Creates or reuses an async httpx client with authentication headers and HTTP/2 support."""
    global _client  # noqa: PLW0603
    if _client is None:
        max_connections = int(os.getenv("SOIL_MAX_CONNECTIONS", "10"))
        _client = httpx.AsyncClient(
            headers={
                "Authorization": f"Bearer {GLOBAL_CONFIG.token!s}",
                "Content-Type": "application/json",
            },
            timeout=30,
            http2=True,
            limits=httpx.Limits(
                max_connections=max_connections,
                max_keepalive_connections=min(20, max_connections),
            ),
        )
    try:
        # This is a workaround to avoid the issue of the token not being updated
        # when the client is reused.
        _client.headers.update(
            {
                "Authorization": f"Bearer {GLOBAL_CONFIG.token!s}",
            }
        )
        yield _client
    finally:
        if _client is not None and _client.is_closed:
            _client = None


async def upload_data(dtype: str, data: Any, metadata: Any) -> Dict[str, Any]:
    """Upload data to the cloud as a new dataset."""
    logger.debug("upload_data:%s", dtype)
    url = API_ROOT + "results/"
    body = {"type": dtype, "data": data, "metadata": metadata}

    async with _get_client() as client:
        response = await client.post(url, json=body)

    if response.status_code != 201:
        raise errors.DataNotUploaded(response.text)
    return response.json()


async def get_result(result_id: str) -> Dict[str, Any]:
    """Get the result data"""
    logger.debug("get_result:%s", result_id)
    url = API_ROOT + "results/" + result_id + "/"

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    return response.json()


async def get_alias(alias: str) -> Dict[str, Any]:
    """Gets an alias"""
    logger.debug("get_alias: %s", alias)
    url = API_ROOT + "states/?name=" + alias + "&with-roles"

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.ObjectNotFound("Error getting alias:" + response.text)

    aliases = response.json()
    if len(aliases) == 0:
        raise errors.DataNotFound("Alias not found")

    alias_id = aliases[0]["_id"]
    url = API_ROOT + "states/" + alias_id + "/"

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.ObjectNotFound("Error getting alias with id " + response.text)
    return response.json()


async def get_result_data(
    result_id: str, query: Dict[str, str] | None = None
) -> Dict[str, Any]:
    """Get the result data"""
    if query is None:
        query = {}
    logger.debug("get_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/data/?" + urlencode(query)

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    return response.json()


async def create_experiment(
    plan: Plan,
    *,
    experiment_group: str | None = None,
    description: dict | str | None = None,
) -> Experiment:
    """Creates an experiment in SOIL asynchronously"""
    logger.debug("create_experiment: %s", str(plan))
    url = API_ROOT + "experiments/"
    experiment = {"name": "", "description": "", "plan": plan}
    if description is not None:
        experiment["description"] = description
    if experiment_group is not None:
        experiment["experiment_group"] = experiment_group

    async with _get_client() as client:
        response = await client.post(url, json={"experiment": experiment})

    if response.status_code != 200:
        raise ValueError("Error creating the experiment:" + response.text)
    return response.json()["experiment"]


async def get_experiment(experiment_id: str) -> Experiment:
    """Gets an experiment from SOIL asynchronously"""
    logger.debug("get_experiment: %s", experiment_id)
    url = API_ROOT + "experiments/" + experiment_id + "/"

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return response.json()


async def get_experiment_logs(experiment_id: str, start_date: str) -> Any:
    """Gets logs from a SOIL experiment asynchronously"""
    logger.debug("get_experiment_id: %s since %s", experiment_id, start_date)
    url = (
        API_ROOT
        + "experiments/"
        + experiment_id
        + "/logs/?"
        + urlencode({"start_date": start_date})
    )

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise ValueError(
            "Error getting the experiment " + experiment_id + "\n" + response.text
        )
    return response.json()


async def set_alias(
    alias: str,
    result_id: str,
    roles: list[str] | None = None,
    extras: dict[str, Any] | None = None,
) -> None:
    """Sets an alias for a result asynchronously. Updates a previous one with the same name."""
    logger.debug("set_alias: %s = %s", alias, result_id)
    obj = {"name": alias, "state": {"alias": alias, "result_id": result_id}}
    if extras is not None:
        obj["state"]["extras"] = extras
    if roles is not None:
        obj["roles"] = roles

    async with _get_client() as client:
        try:
            old_alias = await get_alias(alias)
            old_alias_id = old_alias["_id"]
            url = API_ROOT + "states/" + old_alias_id + "/"
            response = await client.patch(url, json=obj)
        except errors.DataNotFound:
            url = API_ROOT + "states/"
            response = await client.post(url, json=obj)

        if response.status_code != 200:
            raise errors.DataNotUploaded(
                f"Failed to create alias {alias}: {response.text}"
            )

        if roles is not None:
            await _create_app_id_roles(roles=roles)
            url = API_ROOT + "results/" + result_id + "/"
            response = await client.patch(url, json={"roles": roles})
            if response.status_code != 200:
                raise errors.DataNotUploaded(
                    f"Failed to patch result {result_id}: {response.text}"
                )


async def _create_app_id_roles(*, roles: list[str]) -> None:
    api_key = os.environ.get("APP_INFO_API_KEY")
    if api_key is None:
        return

    auth_url = f"{CONF['auth_url']}/api/application/{CONF['auth_app_id']}"
    headers = {"Content-Type": "application/json", "Authorization": api_key}

    async with httpx.AsyncClient(http2=True) as client:
        for role in roles:
            response = await client.get(auth_url, headers=headers)
            if response.status_code != 200:
                logger.debug("Could not retrieve application info.")
                continue

            application_roles: list[dict[Literal["name"], str]] = response.json()[
                "application"
            ]["roles"]
            if len(
                list(
                    takewhile(
                        lambda app_role: app_role["name"] != role, application_roles
                    )
                )
            ) != len(application_roles):
                continue  # role already exists

            response = await client.post(
                f"{auth_url}/role",
                headers=headers,
                json={"role": {"name": role}},
            )
            if response.status_code != 200:
                logger.debug("Role could not be created because %s", response.text)


async def export_result(
    result_id: str, file_path: str, query: Dict[str, str] | None = None
) -> None:
    """Export result and saves it to a file asynchronously"""
    if query is None:
        query = {}
    logger.debug("export_result:%s %s", result_id, str(query))
    url = API_ROOT + "results/" + result_id + "/export/?" + urlencode(query)

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.DataNotFound(response.text)
    with open(file_path, "wb") as fd:
        fd.write(response.content)


class _SoilUploadModule(TypedDict):
    name: str
    code: str
    is_package: bool


async def upload_modules(modules: list[_SoilUploadModule]) -> None:
    """Uploads a chunk of modules asynchronously"""
    logger.debug("upload_modules:%s", modules)
    url = API_ROOT + "modules/"
    data = {"modules": modules}

    async with _get_client() as client:
        response = await client.post(url, json=data)

    if response.status_code != 200:
        raise errors.ModuleNotUploaded(response.text)


async def upload_module(module_name: str, code: str, is_package: bool) -> None:
    """Uploads a module asynchronously"""
    logger.debug("upload_module:%s", module_name)
    url = API_ROOT + "modules/"
    data = {"name": module_name, "code": code, "is_package": is_package}

    async with _get_client() as client:
        response = await client.post(url, json=data)

    if response.status_code != 200:
        raise errors.ModuleNotUploaded(response.text)


async def get_module(full_name: str) -> GetModule:
    """Downloads a module asynchronously"""
    logger.debug("get_module:%s", full_name)
    url = API_ROOT + "modules/" + full_name + "/"

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.ModuleNotFound(response.text)
    return response.json()


async def get_modules() -> list[GetModuleHash]:
    """Gets a list of modules asynchronously"""
    logger.debug("get_modules")
    url = f"{API_ROOT}modules/"
    async with _get_client() as client:
        response = await client.get(url)
    if response.status_code != 200:
        raise errors.ModuleNotFound(response.text)
    return response.json()["result"]


async def create_event(key: str, value: Any) -> Any:
    """Saves an event in soil asynchronously"""
    url = API_ROOT + "alerts/events/"

    async with _get_client() as client:
        response = await client.post(url, json={"key": key, "value": value})

    if response.status_code != 201:
        raise errors.EventNotUploaded("Error saving the event:" + response.text)
    return response.json()


async def create_alert(alert: Dict) -> Any:
    """Creates an alert asynchronously"""
    url = API_ROOT + "alerts/alerts/"

    async with _get_client() as client:
        response = await client.post(url, json=alert)

    if response.status_code != 201:
        raise errors.AlertNotUploaded("Error creating the alert:" + response.text)
    return response.json()


async def get_dictionary(name: str, language: str) -> Dict[str, Any]:
    """Get the dictionary asynchronously"""
    url = API_ROOT + "dictionaries/" + name + "/?language=" + language

    async with _get_client() as client:
        response = await client.get(url)

    if response.status_code != 200:
        raise errors.DictionaryNotFound("Error getting dictionary:" + response.text)
    return response.json()


async def create_dictionary(name: str, language: str, content: Dict) -> Dict[str, Any]:
    """Create a dictionary or update it asynchronously"""
    logger.debug("create_dictionary: %s in language: %s", str(name), str(language))
    data = {
        "name": name,
        "language": language,
        "content": content,
    }
    url = f"{API_ROOT}dictionaries/"

    async with _get_client() as client:
        response = await client.post(
            url,
            content=gzip.compress(bytes(json.dumps(data), encoding="utf-8")),
            headers={"Content-Encoding": "gzip"},
        )

    if response.status_code != 200:
        raise errors.DictionaryNotUploaded(
            f"Error creating/updating the dictionary {name} in language {language}: {response.text}"
        )
    return response.json()


async def get_graph_state(
    case_mix_store_id: str, graph_state_id: str, user_roles: list[str], user_id: str
) -> GraphStateResponse:
    """Get the stored graph state"""
    url = (
        f"{API_ROOT_V3}case-mix-store/{case_mix_store_id}/graph-state/{graph_state_id}"
    )

    async with _get_client() as client:
        response = await client.get(
            url, params={"user_roles": user_roles, "user_id": user_id}
        )

    if response.status_code != 200:
        raise errors.GraphStateNotFound(response.text)
    return response.json()


async def save_graph_state(
    case_mix_store_id: str, request: SaveGraphStateRequest
) -> None:
    """Save the provided graph state"""
    url = f"{API_ROOT_V3}case-mix-store/{case_mix_store_id}/graph-state"

    async with _get_client() as client:
        response = await client.post(url, json=request)

    if response.status_code != 200:
        raise errors.GraphStateNotSaved(response.text)


async def save_pipeline_context(
    context_id: str, subset_id: str, pipeline: PipelineForRequest, user_roles: list[str]
) -> None:
    """Save the provided pipeline context"""
    url = f"{API_ROOT_V3}case-mix-store/{context_id}/pipeline-context"

    async with _get_client() as client:
        response = await client.post(
            url,
            json={
                "subset_id": subset_id,
                "pipeline": pipeline,
                "user_roles": user_roles,
            },
        )

    if response.status_code >= 300:
        raise errors.PipelineContextNotSaved(response.text)


async def get_project_context(
    case_mix_store_id: str, project_id: str, user_roles: list[str]
) -> Context:
    """Get the project context"""
    url = (
        f"{API_ROOT_V3}case-mix-store/{case_mix_store_id}/project-context/{project_id}"
    )

    async with _get_client() as client:
        response = await client.get(url, params={"user_roles": user_roles})

    if response.status_code != 200:
        raise errors.ProjectContextNotFound(response.text)
    return response.json()


async def create_graph_log(
    case_mix_store_id: str,
    graph_state_id: str,
    request: CreateGraphLogRequest,
) -> None:
    """Create a graph log"""
    url = f"{API_ROOT_V3}case-mix-store/{case_mix_store_id}/graph-state/{graph_state_id}/log"

    async with _get_client() as client:
        response = await client.post(url, json=request)

    if response.status_code != 200:
        raise errors.GraphLogNotCreated(response.text)



"""Package that contains calls for soil alias data structures"""

from typing import Any, cast

from soil.aio.api import set_alias
from soil.data_structure import DataStructure
from soil.types import DEFAULT_TIMEOUT, SerializableDataStructure


async def alias(
    name: str,
    data_ref: DataStructure | SerializableDataStructure,
    roles: list[str] | None = None,
    extras: dict[str, Any] | None = None,
    timeout: int = DEFAULT_TIMEOUT,
) -> None:
    """Set an alias to a data reference asynchronously"""
    # Force to get an id if there is none yet.
    data_ref = cast(DataStructure, data_ref)
    if data_ref.id is None:
        await data_ref.async_get_id(timeout=timeout)
    assert isinstance(data_ref.id, str)
    await set_alias(name, data_ref.id, roles, extras)



"""Defines soil.data() async version"""

from typing import (
    Any,
    Dict,
    Optional,
    Type,
    cast,
    overload,
)

from soil import errors
from soil.aio.api import get_alias, get_result, upload_data
from soil.data_structure import DataStructure, get_data_structure_name_and_serialize
from soil.types import DataObject, SerializableDataStructure


@overload
async def data(
    data_object: str,
    metadata: None = None,
    *,
    return_type: None = None,
) -> DataStructure: ...


@overload
async def data[G: SerializableDataStructure](
    data_object: str,
    metadata: None = None,
    *,
    return_type: Type[G],
) -> G: ...


@overload
async def data[D: DataObject](
    data_object: D,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: None = None,
) -> DataStructure[D]: ...


@overload
async def data[G: SerializableDataStructure](
    data_object: DataObject,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: Type[G],
) -> G: ...


async def data[D: DataObject, G: SerializableDataStructure](  # pyright: ignore[reportInconsistentOverload]
    data_object: str | D,
    metadata: dict[str, Any] | None = None,
    *,
    return_type: Type[G] | None = None,
) -> G | DataStructure | DataStructure[D]:
    """Load data from the cloud or mark it as uploadable"""
    cast_return_type = DataStructure if return_type is None else return_type
    if isinstance(data_object, str):
        # Data object is an id or an alias
        try:
            data_object = await _load_data_alias(data_object)
        except errors.DataNotFound:
            pass
        return cast(cast_return_type, await _load_data_id(data_object))  # pyright: ignore
    return cast(
        cast_return_type,  # pyright: ignore
        await _upload_data(data_object, metadata),
    )


async def _upload_data(
    data_object: Any, metadata: Optional[Dict[str, Any]] = None
) -> DataStructure:
    ds_name, serialized = get_data_structure_name_and_serialize(data_object)
    result = await upload_data(ds_name, serialized, metadata)
    ds = DataStructure(result["_id"], dstype=result["type"])
    return ds


async def _load_datastructure(did: str, dtype: str) -> DataStructure:
    # TODO: dynamically load a data structure
    return DataStructure(did, dstype=dtype)


async def _load_data_alias(alias: str) -> str:
    result = await get_alias(alias)
    return result["state"]["result_id"]


async def _load_data_id(data_id: str) -> DataStructure:
    result = await get_result(data_id)
    return await _load_datastructure(result["_id"], result["type"])



import unittest
from unittest.mock import patch

from soil import errors
from soil.aio.data import DataStructure, data
from soil.types import SerializableDataStructure


class TestData(unittest.IsolatedAsyncioTestCase):
    @patch("soil.aio.data.get_alias")
    @patch("soil.aio.data.get_result")
    async def test_data_with_string_id(self, mock_get_result, mock_get_alias):
        mock_get_alias.side_effect = errors.DataNotFound()
        mock_get_result.return_value = {"_id": "test_id", "type": "test_type"}
        result = await data("test_id")
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "test_id")
        self.assertEqual(result.dstype, "test_type")
        mock_get_alias.assert_called_once_with("test_id")
        mock_get_result.assert_called_once_with("test_id")

    @patch("soil.aio.data.get_alias")
    @patch("soil.aio.data.get_result")
    async def test_data_with_alias(self, mock_get_result, mock_get_alias):
        mock_get_alias.return_value = {"state": {"result_id": "resolved_id"}}
        mock_get_result.return_value = {"_id": "resolved_id", "type": "test_type"}
        result = await data("test_alias")
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "resolved_id")
        self.assertEqual(result.dstype, "test_type")

    @patch("soil.aio.data.upload_data")
    async def test_data_with_data_object(self, mock_upload_data):
        test_data = {"key": "value"}
        mock_upload_data.return_value = {"_id": "new_id", "type": "dict"}
        result = await data(test_data)
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "new_id")
        self.assertEqual(result.dstype, "dict")

    @patch("soil.aio.data.get_alias")
    @patch("soil.aio.data.get_result")
    async def test_data_with_return_type(self, mock_get_result, mock_get_alias):
        class TestType(SerializableDataStructure):
            pass

        mock_get_alias.side_effect = errors.DataNotFound()
        mock_get_result.return_value = {"_id": "test_id", "type": "test_type"}
        result = await data("test_id", return_type=TestType)
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "test_id")  # pyright:ignore[reportAttributeAccessIssue]
        self.assertEqual(result.dstype, "test_type")  # pyright:ignore[reportAttributeAccessIssue]

    @patch("soil.aio.data.upload_data")
    async def test_data_with_metadata(self, mock_upload_data):
        test_data = {"key": "value"}
        test_metadata = {"meta_key": "meta_value"}
        mock_upload_data.return_value = {"_id": "new_id", "type": "dict"}
        result = await data(test_data, metadata=test_metadata)
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "new_id")
        self.assertEqual(result.dstype, "dict")
        mock_upload_data.assert_called_once()
        # Check if metadata is passed correctly
        args, _ = mock_upload_data.call_args
        self.assertEqual(args[2], test_metadata)

    @patch("soil.aio.data.get_alias")
    @patch("soil.aio.data.get_result")
    async def test_data_not_found(self, mock_get_result, mock_get_alias):
        mock_get_alias.side_effect = errors.DataNotFound()
        mock_get_result.return_value = {"_id": "test_id", "type": "test_type"}
        result = await data("non_existent_alias")
        self.assertIsInstance(result, DataStructure)
        self.assertEqual(result.id, "test_id")
        self.assertEqual(result.dstype, "test_type")


if __name__ == "__main__":
    unittest.main()



"""Exceptions for SOIL-SDK"""


class SoilSDKError(Exception):
    """Common soil-sdk exception"""


class InternalSoilError(SoilSDKError):
    """Generic soil-sdk error"""


class LoginError(SoilSDKError):
    """Exception to raise when unable to login Soil"""


class ExperimentError(SoilSDKError):
    """Exception to raise when an error has occurred while executing a Soil experiment"""


class ExperimentTimeoutError(SoilSDKError):
    """Exception to raise when a Soil experiment times out"""


class ObjectNotFound(SoilSDKError):
    """Exception to raise when Soil object could not be found"""


class DictionaryNotFound(ObjectNotFound):
    """Exception to raise when Soil Dictionary could not be found"""


class GraphStateNotFound(ObjectNotFound):
    """Exception to raise when Soil Graph State could not be found"""


class DataNotFound(ObjectNotFound):
    """Exception to raise when Soil data could not be found"""


class JobCollectionNotFound(ObjectNotFound):
    """Exception to raise when Soil job collection could not be found"""


class ModuleNotFound(ObjectNotFound):
    """Exception to raise when Soil module could not be found"""


class ObjectNotUploaded(SoilSDKError):
    """Exception to raise when Soil object could not be uploaded"""


class GraphStateNotSaved(ObjectNotUploaded):
    """Exception to raise when Soil Graph State could not be saved"""


class DictionaryNotUploaded(ObjectNotUploaded):
    """Exception to raise when Soil Dictionary could not be uploaded"""


class DataNotUploaded(ObjectNotUploaded):
    """Exception to raise when Soil data could not be uploaded"""


class ModuleNotUploaded(ObjectNotUploaded):
    """Exception to raise when Soil Module could not be uploaded"""


class AlertDataNotUploaded(ObjectNotUploaded):
    """Exception to raise when Soil alert could not be uploaded"""


class AlertNotUploaded(AlertDataNotUploaded):
    """Exception to raise when Soil alert condition could not be uploaded"""


class EventNotUploaded(AlertDataNotUploaded):
    """Exception to raise when Soil event alert could not be uploaded"""


class DataStructureError(SoilSDKError):
    """Exception to raise when Soil DataStructure has any error"""


class DataStructureType(DataStructureError):
    """Exception to raise when Soil DataStructure type is not recognised"""


class DataStructurePipelineNotFound(DataStructureError):
    """Exception to raise when Soil DataStructure Pipeline is not found"""


class PipelineContextNotSaved(DataStructureError):
    """Exception to raise when Soil Pipeline is not found"""


class ProjectContextNotFound(DataStructureError):
    """Exception to raise when Soil Project Context is not found"""


class GraphLogNotCreated(SoilSDKError):
    """Exception to raise when Soil Graph Log could not be created"""



"""Module that defines some useful functions"""

from typing import Any, Callable, Dict, List
from uuid import uuid4

from pandas import Timestamp


def _isalambda(value: Any) -> bool:
    return callable(value) and value.__name__ == "<lambda>"


def generate_data_structure_ids(fn_name: str, quantity: int) -> List[str]:
    """Generates an id for a datastructure"""
    rnd = str(uuid4())
    return [
        fn_name.replace(".", "/") + "-" + str(rnd) + "-" + str(i)
        for i in range(quantity)
    ]


def build_function(function: Callable) -> Dict[str, Any]:
    """Mounts the argument structure to serialize a function. It uses the reserved dict key
    `__soil_arg_type` to mark a function. The function code must be inside modules or
    data_structures folders.
    It recognises 3 types of functions:
        * Named functions: `__soil_arg_type='function'` functions with a module
            and a name.
        * Lambda functions: `__soil_arg_type='lambda'` They are not actually serialized
            but kept as a placeholder for the pattern matching in the server.
            They can only be used as a decorator parameter.
        * Decorated functions: `__soil_arg_type='decorated'` A function that has been
            decorated with one or more decorators that have been decorated with
            soil.decorator or functions returned by that decorated function.
    """
    first_qual = function.__qualname__.split(".")[0]
    if function.__module__ == "soil.decorator" and first_qual == "decorator":
        wrapped, calls = function(__show_calls=True)
        return {
            "__soil_arg_type": "decorated",
            "module": wrapped.__module__,
            "name": wrapped.__name__,
            "calls": [
                {
                    "args": [build_arguments(arg) for arg in call[0]],
                    "kwargs": build_arguments(call[1]),
                }
                for call in calls
            ],
        }

    if _isalambda(function):
        return {"__soil_arg_type": "lambda"}

    return {
        "__soil_arg_type": "function",
        "module": function.__module__,
        "name": function.__name__,
    }


def build_arguments(args: Any) -> Any:
    """Transforms the module arguments into serializable arguments"""
    if isinstance(args, dict):
        return {key: build_arguments(value) for (key, value) in args.items()}

    if isinstance(args, list):
        return [build_arguments(value) for value in args]

    if callable(args):
        return build_function(args)

    if isinstance(args, Timestamp):
        additional_kwargs = {}
        if (
            args.tzinfo
            and hasattr(args.tzinfo, "zone")
            and args.tzinfo.zone is not None
        ):
            additional_kwargs = {"tz": args.tzinfo.zone}
        return {
            "_object": Timestamp.__name__,
            "_constructor_params": {
                "positional_args": [args.isoformat()],
                "keywords_args": {**additional_kwargs},
            },
        }

    return args


def generate_transformation(
    input_ids: List[str], output_ids: List[str], fn_name: str, args: Dict[str, Any]
) -> Dict[str, Any]:
    """Mounts the dictionary for a module in a plan"""
    tid = fn_name + "-" + str(uuid4())

    return {
        "id": tid,
        "module": fn_name,
        "inputs": input_ids,
        "outputs": output_ids,
        "args": build_arguments(args),
    }



"""Module for Object Storage"""

from dataclasses import dataclass
from typing import Any, Dict, Optional

from soil.storage.base_storage import BaseStorage


@dataclass
class ObjectStorage(BaseStorage):
    """Implements basic Object Storage using s3 api. If there are no s3 credentials
    it fallsback to disk. Before storing the data they are compressed with zlib.

    Example:
    -------
        class ObjectDS(DataStructure):
        def serialize(self):
            obj_storage = ObjectStorage()
            obj_storage.put_object(json.dumps(self.data).encode('utf-8'))
            return obj_storage

        @staticmethod
        def deserialize(obj_storage: ObjectStorage, metadata):
            raw_data = obj_storage.get_object()
            data = json.loads(raw_data.decode('utf-8'))
            return ObjectDS(data, metadata, storage=obj_storage)

        def get_data(self):
            return self.data

    Attributes:
    ----------
        path: str = '': An optional folder to store the data to.
        obj_name: Optional[str]: The obj_name can optionally be provided. Otherwise
            a unique name will be generated. Caution! Existing data for the same app
            will be overwritten if the same name is used twice.
        metadata: Optional[Dict[str, Any]]: Experimental. Metadata to be stored in the s3 storage.

    """

    path: str = ""
    obj_name: Optional[str] = None
    metadata: Optional[Dict[str, Any]] = None

    def put_object(self, obj: bytes) -> None:
        """Compressess an object and stores it."""

    def get_object(self) -> bytes:
        """Gets the object back and decompresses it."""



"""Package for soil.task"""

from typing import Any, Callable, List

from soil.data_structure import DataStructure


def task(modulified_module: Callable) -> Callable:
    """Decorates a modulified module in soil. This function is to call
    soil modules from other soil modules.

    Example:
    -------
        task(my_module)(data, arg1='val1')

    """

    def decorator(*args: DataStructure, **kwargs: Any) -> List[DataStructure]:
        return modulified_module(*args, **kwargs)

    return decorator


def task_wait(futures: Any) -> Any:
    """Wait until computation completes and gather results.

    Accepts a future, nested container of futures, iterator, or queue.
    The return type will match the input type.

    Example:
    -------
        result = tasks_wait([future1, future2])

    """
    return futures



import unittest
from unittest.mock import AsyncMock, patch

from soil import errors
from soil.aio.alias import alias
from soil.data_structure import DataStructure
from soil.types import DEFAULT_TIMEOUT


class TestAlias(unittest.IsolatedAsyncioTestCase):
    async def test_alias_1(self):
        # Mock dependencies
        mock_data_ref = AsyncMock(spec=DataStructure)
        mock_data_ref.id = None

        def mock_set_id(**kwargs):
            mock_data_ref.id = "test_id"
            return mock_data_ref.id

        mock_data_ref.async_get_id.side_effect = mock_set_id

        with patch(
            "soil.aio.alias.set_alias", new_callable=AsyncMock
        ) as mock_set_alias:
            # Test alias function
            await alias(
                "test_name",
                mock_data_ref,
                roles=["role1", "role2"],
                extras={"key": "value"},
            )

            # Assert that async_get_id was called
            mock_data_ref.async_get_id.assert_called_once_with(timeout=DEFAULT_TIMEOUT)

            # Assert that set_alias was called with correct arguments
            mock_set_alias.assert_called_once_with(
                "test_name", "test_id", ["role1", "role2"], {"key": "value"}
            )

    async def test_alias_with_existing_id(self):
        # Mock dependencies
        mock_data_ref = AsyncMock(spec=DataStructure)
        mock_data_ref.id = "existing_id"

        with patch(
            "soil.aio.alias.set_alias", new_callable=AsyncMock
        ) as mock_set_alias:
            # Test alias function
            await alias("test_name", mock_data_ref)

            # Assert that async_get_id was not called
            mock_data_ref.async_get_id.assert_not_called()

            # Assert that set_alias was called with correct arguments
            mock_set_alias.assert_called_once_with(
                "test_name", "existing_id", None, None
            )

    async def test_alias_with_invalid_data_ref(self):
        mock_data_ref = DataStructure()
        with self.assertRaises(errors.DataStructurePipelineNotFound):
            await alias("test_name", mock_data_ref)


if __name__ == "__main__":
    unittest.main()



"""Setup configuration for soil-sdk"""

from setuptools import setup

if __name__ == "__main__":
    setup(
        use_scm_version={
            "write_to": "soil/_version.py",
            "write_to_template": '__version__ = "{version}"',
        },
    )



"""Defines the configuration for Soil"""

import json
import logging
import os
import time
from os import getenv
from typing import NamedTuple, Optional

import jwt
import requests

from soil import errors

# TODO add windows support

# pylint: disable=invalid-name
env = getenv("PY_ENV", "development")


def _refresh_token(auth_host: str, refresh_token: str) -> str:
    url = auth_host + "/api/jwt/refresh"
    response = requests.post(url, json={"refreshToken": refresh_token}, timeout=30)
    if response.status_code != 200:
        raise errors.LoginError("Invalid refresh token please run soil login again.")
    return json.loads(response.text)["token"]


def get_soil_root(relpath: str) -> Optional[str]:
    """Checks if the current dir is under a soil environment
    and returns its root. Returns None otherwise.
    """
    path = os.path.abspath(relpath) + "/"
    while path != "/":
        path, _ = os.path.split(path)
        if "soil.yml" in os.listdir(path):
            return path
    return None


project_root = get_soil_root(".")
SOIL_URL = ""
if env != "test":
    try:
        if project_root is None:
            raise FileNotFoundError("Project root not found.")
        with open(project_root + "/soil.conf", encoding="utf-8") as conf_file:
            CONF = json.loads(conf_file.read())
            SOIL_URL = CONF["soil_url"]
        logging.info(
            "Loaded config from %s using app_id %s",
            project_root + "/soil.conf",
            CONF["auth_app_id"],
        )
    except FileNotFoundError:
        try:
            with open(
                getenv("HOME", "") + "/.soil/soil.conf", encoding="utf-8"
            ) as conf_file:
                CONF = json.loads(conf_file.read())
                SOIL_URL = CONF["soil_url"]
            logging.warning(
                "Loaded config from %s using app_id %s. It is recommended to have a soil.conf file at the root of the project.",  # pylint:disable=line-too-long
                getenv("HOME", "") + "/.soil/soil.conf",
                CONF["auth_app_id"],
            )

        except FileNotFoundError:
            logging.error(
                "~/.soil/soil.conf file not found. Please run soil configure."
            )
else:
    SOIL_URL = "http://test_host.test"
    CONF = {
        "auth_api_key": "test_auth_api_key",
        "auth_app_id": "test_auth_app_id",
        "auth_url": "test_auth_url",
        "soil_url": "test_soil_url",
    }
    os.environ["APP_INFO_API_KEY"] = "TEST_APP_INFO_API_KEY"

TOKEN = ""  # nosec

if env != "test":
    try:
        with open(
            getenv("HOME", "") + "/.soil/" + CONF["auth_app_id"] + "/soil.env",
            encoding="utf-8",
        ) as env_file:
            ENV = json.loads(env_file.read())
            TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
    except FileNotFoundError:
        logging.warning(
            (
                "~/.soil/%s/soil.env file not found."
                " Using ~/.soil/soil.env instead. It is recommended to run soil login."
            ),
            CONF["auth_app_id"],
        )
        try:
            with open(
                getenv("HOME", "") + "/.soil/soil.env", encoding="utf-8"
            ) as env_file:
                ENV = json.loads(env_file.read())
                TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
        except FileNotFoundError:
            logging.error("~/.soil/soil.env file not found. Please run soil login.")
else:
    TOKEN = "test_token"  # nosec

DEFAULT_CONFIG = {
    "host": getenv("SOIL_HOST", SOIL_URL),
    "token": getenv("SOIL_TOKEN", TOKEN),
}

if env != "test":
    decoded_token = jwt.decode(TOKEN, options={"verify_signature": False})
    conf_app_id = CONF["auth_app_id"].strip()
    token_app_id = decoded_token["applicationId"]

    if token_app_id != conf_app_id:
        raise errors.LoginError(
            "Application Id for the project config "
            + f"{conf_app_id} does not coincide with the token application id {token_app_id}.\n"
            + "Please, run 'soil login' again."
        )


class SoilConfiguration(NamedTuple):
    """Soil configuration class"""

    host: Optional[str]
    token: Optional[str]

    def __getattribute__(self, name: str) -> str:
        """Refresh the token if necessary before accessing it."""
        # pylint:disable=global-statement
        global TOKEN  # noqa: PLW0603

        if name != "token":
            return object.__getattribute__(self, name)

        if env == "test":
            return TOKEN

        # Min nummber of seconds before expire that will trigger a refresh.
        min_delta = 10

        exp = jwt.decode(TOKEN, options={"verify_signature": False})["exp"]
        if exp - time.time() < min_delta:
            TOKEN = _refresh_token(CONF["auth_url"], ENV["auth"]["refreshToken"])
        return TOKEN


GLOBAL_CONFIG = SoilConfiguration(**DEFAULT_CONFIG)


# Not used for now
# def config(token: Optional[str] = None, host: Optional[str] = None) -> SoilConfiguration:
#     ''' Set the Soil's configuration '''
#     global GLOBAL_CONFIG  # pylint: disable=global-statement
#     new_config = SoilConfiguration(token=token, host=host)
#     GLOBAL_CONFIG = SoilConfiguration(**{**GLOBAL_CONFIG._asdict(), **new_config._asdict()})
#     return GLOBAL_CONFIG
